/*
 * Pouze mono wav soubory!
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <getopt.h>
#include <unistd.h>
#include <math.h>
#include <ctype.h>
#include <inttypes.h>

#include <fftw3.h>
#include <sndfile.h>

#define HOP_TIME 0.1
#define FFT_TIME 1

int help() {
    printf("usage: -i audio.wav -s sample.wav -p [0-100]\n");

    return 0;
}

static double audio_get_duration (SF_INFO *sfinfo) {
    if (sfinfo->samplerate < 1) return 0;

    if (sfinfo->frames / sfinfo->samplerate > 0x7FFFFFFF) return -1;

    return ((1.0 * sfinfo->frames) / sfinfo->samplerate);
}

static void audio_print_info (char *name, SF_INFO *sfinfo) {
    printf("file name     : %s\n", name);
    printf("duration      : %.3f s\n",audio_get_duration(sfinfo));
    printf("frames        : %" PRId64 "\n", sfinfo->frames);
    printf("channels      : %d\n", sfinfo->channels);
    printf("sample rate   : %d Hz\n", sfinfo->samplerate);
    printf("format        : 0x%08X\n", sfinfo->format);
}

static int fft_freq_map (int fft_size, int sample_rate, int *freq_map) {
    double d = sample_rate / fft_size;
    int i = (int) (2 / (pow(2, 0.08333333333333333) - 1));
    int j = (int) round(log(i * d / 440) / log(2) * 12 + 69);
    int k = 0;

    while (k <= i) {
        /* nelze k++, jinak si kompilátor stěžuje */
        freq_map[k] = k + 1;
        k++;
    }

    while (k <= (fft_size / 2)) {
        double d2 = log(k * d / 440) / log(2) * 12 + 69;
        if (d2 > 127) d2 = 127;
        freq_map[(k++)] = (i + (int) round(d2) - j);
    }

    return freq_map[(k - 1)] + 1;
}

static void fft_hamming (double *array, int length) {
    int i = 0, k;
    double d1 = (1.0 / length / 0.54);
    double d2 = 6.283185307179586 / length;

    for (k = 0; i < length; ++k) {
        array[k] = (d1 * (0.5434782608695652 - (0.4565217391304348 * cos(d2 * k))));
        i++;
    }
}

int main(int argc, char **argv) {
    int         i = 0, j = 0;
    int         limit = 70;
    int         readcount;
    int         sample_rate = 0;
    char        *in_file = NULL, *sp_file = NULL;
    SNDFILE     *input = NULL;
    SNDFILE     *sample = NULL;
    SF_INFO     in_file_info, sp_file_info;

    while ((i = getopt(argc, argv, "+i:s:p:h")) != EOF) {
        switch (i) {
            case 'i':
                in_file = strdup(optarg);
                break;
            case 's':
                sp_file = strdup(optarg);
                break;
            case 'p':
                limit = atoi(optarg);
                break;
            case 'h':
            default:
                help();
                break;
        }
    }

    /* kontrola na zadané soubory */
    if ((in_file == NULL) || (sp_file == NULL)) {
        help();
        return 1;
    }

    /* otevřeme hlavní soubor */
    if ((input = sf_open(in_file, SFM_READ, &in_file_info)) == NULL) {
        printf("Can't open input file: %s\n", in_file);
        return 1;
    }

    /* načteme vzorkový soubor */
    if ((sample = sf_open(sp_file, SFM_READ, &sp_file_info)) == NULL) {
        printf("Can't open sample file: %s\n", sp_file);
        return 1;
    }

    /* zobrazíme informace o souborech */
    audio_print_info(in_file, &in_file_info);
    printf("\n");
    audio_print_info(sp_file, &sp_file_info);
    printf("\n");

    if (in_file_info.samplerate != sp_file_info.samplerate) {
        printf("Input sample-rate is not equal to sample file!");
        return 1;
    }

    /* nastavíme vzorkovací frekvenci */
    sample_rate = sp_file_info.samplerate;

    /* hop_size a fft_size */
    int hop_size = (int) sample_rate * HOP_TIME;
    int fft_size = (int) pow(2,round(log(FFT_TIME * sample_rate) / log(2)));

    /* frekvenční mapa */
    int freq_map[((fft_size / 2) + 1)];

    /* kruhový buffer */
    double circ_buffer[fft_size];
    int circ_buffer_index = 0;

    /* nastavení frekvenční mapy, můžeme až po definici kruhového bufferu! */
    int freq_map_size = fft_freq_map(fft_size, sample_rate, freq_map);

    /* fft okno */
    double window[fft_size];

    /* zajímá nás pouze hammingův filtr */
    fft_hamming(window, fft_size);

    for (i = 0; i < fft_size; i++) {
        window[i] *= sqrt(fft_size);
    }

    printf("Audio spacing: %d\n", hop_size);
    printf("FFT size: %d\n", fft_size);

    /* velikost bufferu pro načítání audia */
    int buffer_len = hop_size * sp_file_info.channels;
    float buffer[buffer_len];

    /* pokud je buffer větší než velikost fft tak končíme */
    if (buffer_len > fft_size) {
        printf("Input buffer (%d) > circute buffer (%d)!\n", buffer_len, fft_size);
        return 1;
    }

    /* definice pro libfftw3 */
    fftw_complex *fft_in, *fft_out;
    fftw_plan p;

    /* alokace paměti pro fft_in a fft_out */
    fft_in = (fftw_complex*) fftw_malloc(sizeof(fftw_complex) * fft_size);
    fft_out = (fftw_complex*) fftw_malloc(sizeof(fftw_complex) * fft_size);

    /* nastavení plánu fft */
    p = fftw_plan_dft_1d(fft_size, fft_in, fft_out, FFTW_FORWARD, FFTW_ESTIMATE);

    /* rezervujeme paměť pro framy vzorku */
    double frames[(int) (ceil(sp_file_info.frames / buffer_len) + 1)][freq_map_size];
    int frame_number = 0;

    /* načítání dat z vzorového audio souboru */
    while ((readcount = sf_read_float(sample, buffer, buffer_len))) {
        for (i = 0; i < readcount; i++) {
            /* pozor: float -> double */
            circ_buffer[circ_buffer_index++] = buffer[i];

            /* začneme od začátku */
            if (circ_buffer_index >= fft_size) circ_buffer_index = 0;
        }

        /* buffer uložíme do pole pro fft */
        for (i = 0; i < fft_size; i++) {
            /* aplikujeme fft okno */
            fft_in[i][0] = circ_buffer[circ_buffer_index++] * window[i];

            /* imaginární část je nulová */
            fft_in[i][1] = 0;

            /* opět začneme od začátku kruhového bufferu */
            if (circ_buffer_index >= fft_size) circ_buffer_index = 0;
        }

        /* spustíme fftw */
        fftw_execute(p);

        /* Musíme konkrétní frame předem vynulovat */
        for (i = 0; i < freq_map_size; i++) {
            frames[frame_number][i] = 0;
        }

        /* Použijeme frekvenční mapu */
        for (i = 0; i < ((fft_size / 2) + 1); i++) {
            //printf("%10.10f\n",pow(fft_out[i][0],2) + pow(fft_out[i][1],2));
            //printf("%d -> %d => %.10f = %.10f\n", i, freq_map[i], frame[freq_map[i]], (pow(fft_out[i][0],2) + pow(fft_out[i][1],2)));
            frames[frame_number][freq_map[i]] += pow(fft_out[i][0],2) + pow(fft_out[i][1],2);
        }

        /* zobrazíme frekvenční mapu */
        //for (i = 0; i < freq_map_size; i++) {
        //    printf("%.10f\n",frames[frame_number][i]);
        //}

        frame_number++;
    }

    /* zavřeme soubor */
    sf_close(sample);

    int frames_size = (frame_number - 1);

    printf("\n");

    /*************************** hlavní audio soubor ***********************/

    /* vynulujeme pozici */
    frame_number = 0;

    /* vynulujeme kruhový buffer */
    circ_buffer_index = 0;
    for (i = 0; i < fft_size; i++) {
        circ_buffer[i] = 0;
    }

    /* matche jednotlivých pozic */
    int frames_match_size = (int) (ceil(in_file_info.frames / buffer_len) + 1);
    int frames_match[frames_match_size];
    for (i = 0; i < frames_match_size; i++) {
        frames_match[i] = 0;
    }

    /* začneme porovnávání s reálným souborem */
    while ((readcount = sf_read_float(input, buffer, buffer_len))) {
        for (i = 0; i < readcount; i++) {
            /* pozor: float -> double */
            circ_buffer[circ_buffer_index++] = buffer[i];

            /* začneme od začátku */
            if (circ_buffer_index >= fft_size) circ_buffer_index = 0;
        }

        /* buffer uložíme do pole pro fft */
        for (i = 0; i < fft_size; i++) {
            /* aplikujeme fft okno */
            fft_in[i][0] = circ_buffer[circ_buffer_index++] * window[i];

            /* imaginární část je nulová */
            fft_in[i][1] = 0;

            /* opět začneme od začátku kruhového bufferu */
            if (circ_buffer_index >= fft_size) circ_buffer_index = 0;
        }

        /* spustíme fftw */
        fftw_execute(p);

        /* definice frame */
        double frame[freq_map_size];
        for (i = 0; i < freq_map_size; i++) {
            frame[i] = 0;
        }

        /* Použijeme frekvenční mapu */
        for (i = 0; i < ((fft_size / 2) + 1); i++) {
            frame[freq_map[i]] += pow(fft_out[i][0],2) + pow(fft_out[i][1],2);
        }

        //for (i = 0; i < freq_map_size; i++) {
        //    printf("%.10f\n",frame[i]);
        //}

        double similarity = 0;
        double similarity_max = 0;
        int similarity_max_index = 0;
        double real_position = 0;

        /* najdeme nejlépe odpovídající vzorky ve frames */
        for (i = 0; i < frames_size; i++) {
            similarity = 0;
            /* cca 120 vzorků */
            for (j = 0; j < freq_map_size; j++) {
                double diff;
                /* ochrana proti dělení nulou */
                if (frame[j] == frames[i][j]) {
                    diff = 1;
                } else if (frame[j] == 0) {
                    diff = (frames[i][j] / frame[j]);
                } else {
                    diff = (frames[i][j] / frame[j]);
                }
                /* nahoře byl větší, takže musíme výraz obrátit */
                if (diff > 1) diff = 1 / diff;

                //printf ("    block %d, old %.10f, new %.10f, diff %.10f\n", j,  frames[i][j], frame[j], diff);

                similarity += diff;
            }
            similarity /= freq_map_size;
            //if (similarity_max, similarity / freq_map_size)
            if (similarity_max < similarity) {
                similarity_max = similarity;
                similarity_max_index = i;
            }
        }

        similarity_max *= 100;

        if (similarity_max >= limit) {
            /* vypočítáme předpokládanou pozici */
            real_position = ((frame_number - similarity_max_index) * buffer_len / (double) sample_rate);
            frames_match[(frame_number - similarity_max_index)]++;
            //printf ("similarity %.10f, freq_map_size %d\n", similarity, freq_map_size);
            printf ("Match on position: %5.3f, sample position: %5.3f, similarity: %3.2f, real position: %5.3f\n",  (frame_number * buffer_len / (double) sample_rate), (similarity_max_index * buffer_len / (double) sample_rate), similarity_max,real_position);
        }

        frame_number++;
    }

    /* zavřeme soubor */
    sf_close(input);


    /*************************** zobrazíme výsledky ************************/

    printf("\n");
    printf("Final results:\n");

    for (i = 0; i < frames_match_size; i++) {
        if (frames_match[i] < 1) continue;
        printf ("Position: %5.3f seconds, weight: %d\n", (i * buffer_len / (double) sample_rate), frames_match[i]);
    }

    printf("\n");

    /********************************** konec ******************************/

    /* ukončíme fft */
    fftw_destroy_plan(p);
    fftw_free(fft_in);
    fftw_free(fft_out);

    return 0;
}
